use quote::quote;
use syn::Field;

use crate::{extract_inner_type, is_one_of_type, is_type};

/// Generate the ChangeBuilder field fns
pub fn builder_change_fields(field: &Field) -> proc_macro2::TokenStream {
    let field_name = &field.ident.as_ref().unwrap();
    let field_type = &field.ty;
    let field_name_str = field_name.to_string();

    // Never update _id
    if field_name_str == "_id" {
        return quote! {};
    }

    let number_types = [
        "u8", "u16", "u32", "u64", "u128", "usize", "i8", "i16", "i32", "i64", "i128", "isize",
        "f16", "f32", "f64", "f128",
    ];

    // Number type fn
    if is_one_of_type(field_type, &number_types) {
        let documentation = format!("Set the value of `{field_name}`");
        let inc_fn_name = syn::Ident::new(&format!("{}_increment", field_name), field_name.span());
        let doc_inc = format!("Increment value of `{field_name}` by `value`. Consecutive calls to this function will not add up, they overwrite the increment.");

        let mul_fn_name = syn::Ident::new(&format!("{}_multiply", field_name), field_name.span());
        let doc_mul = format!("Multiply value of `{field_name}` by `value`. Consecutive calls to this function will not add up, they overwrite the multiply.");

        return quote! {
        #[doc = #documentation]
        pub fn #field_name(mut self, value: #field_type)-> Self {
            self.model.#field_name = value.into();

            self.changeset.entry("$set".to_string()).or_insert(mongod::mongodb::bson::doc! {}.into()).as_document_mut().unwrap().insert(
                #field_name_str.to_string(),
                mongod::mongodb::bson::to_bson(&self.model.#field_name).unwrap(),
            );

            self
        }

        #[doc = #doc_inc]
        pub fn #inc_fn_name(mut self, value: #field_type) -> Self {
            self.model.#field_name += value;

            self.changeset.entry("$inc".to_string()).or_insert(mongod::mongodb::bson::doc! {}.into()).as_document_mut().unwrap()
            .insert(
                #field_name_str.to_string(),
                mongod::mongodb::bson::to_bson(&value).unwrap(),
            );

            self
        }

        #[doc = #doc_mul]
        pub fn #mul_fn_name(mut self, value: #field_type) -> Self {
            self.model.#field_name *= value;

            self.changeset.entry("$mul".to_string()).or_insert(mongod::mongodb::bson::doc! {}.into()).as_document_mut().unwrap()
            .insert(
                #field_name_str.to_string(),
                mongod::mongodb::bson::to_bson(&value).unwrap(),
            );

            self
        }
        };
    }

    if is_type(field_type, "Vec") {
        let inner_field_type = extract_inner_type(field_type, "Vec").unwrap();

        let push_fn_name = syn::Ident::new(&format!("{}_push", field_name), field_name.span());

        let documentation = format!("Add a value to the Vec `{field_name}`");
        let documentation2 = format!("Set the value of `{field_name}`");

        return quote! {
        #[doc = #documentation]
        pub fn #push_fn_name<T>(mut self, value: T) -> Self where T: Into<#inner_field_type> + serde::Serialize {
            let mut push = self.changeset.entry("$push".to_string()).or_insert(mongod::mongodb::bson::doc! {}.into()).as_document_mut().unwrap();

            if push.contains_key(#field_name_str) {
                let current = push.get_mut(#field_name_str.to_string()).unwrap();

                if current.as_document().map(|x| !x.contains_key("$each")).unwrap_or(true) {
                    let each = mongod::mongodb::bson::doc! {
                        "$each": [current, mongod::mongodb::bson::to_bson(&value).unwrap()]
                    };

                    push.insert(#field_name_str.to_string(), each);
                } else {
                    current.as_document_mut().unwrap().get_mut("$each").unwrap().as_array_mut().unwrap().push(mongod::mongodb::bson::to_bson(&value).unwrap());
                }
            } else {
                push.insert(#field_name_str.to_string(), mongod::mongodb::bson::to_bson(&value).unwrap());
            }

            self.model.#field_name.push(value.into());

            self
        }

        #[doc = #documentation2]
        pub fn #field_name<T>(mut self, value: T)-> Self where T: Into<#field_type> + serde::Serialize  {
            self.model.#field_name = value.into();

            self.changeset.entry("$set".to_string()).or_insert(mongod::mongodb::bson::doc! {}.into()).as_document_mut().unwrap().insert(
                #field_name_str.to_string(),
                mongod::mongodb::bson::to_bson(&self.model.#field_name).unwrap(),
            );

            self
        }
        };
    }

    if is_type(field_type, "Historic") {
        let inner_field_type = extract_inner_type(field_type, "Historic").unwrap();

        let documentation = format!(
            "Update the value of `{field_name}`. This change will be recorded by the `Historic`"
        );

        // Code for Historic<T>
        return quote! {
            #[doc = #documentation]
            pub fn #field_name<T>(mut self, value: T) -> Self where T: Into<#inner_field_type> + serde::Serialize {
                self.model.#field_name.update(value.into());

                self.changeset.entry("$set".to_string()).or_insert(mongod::mongodb::bson::doc! {}.into()).as_document_mut().unwrap().insert(
                    #field_name_str.to_string(),
                    mongod::mongodb::bson::to_bson(&self.model.#field_name).unwrap(),
                );


                self
            }
        };
    }

    if is_type(field_type, "Option") {
        let inner_field_type = extract_inner_type(field_type, "Option").unwrap();

        if is_type(inner_field_type, "Historic") {
            let inner_field_type = extract_inner_type(inner_field_type, "Historic").unwrap();

            let documentation = format!("Update the value of `{field_name}`. This change will be recorded by the `Historic`. If `{field_name}` is `None` a new `Historic` will be initialized.");

            // Code for Option<Historic<T>>
            return quote! {
                #[doc = #documentation]
                pub fn #field_name<T>(mut self, value: T) -> Self where T: Into<#inner_field_type> + serde::Serialize {
                    if let Some(mut opt) = self.model.#field_name.as_mut() {
                        opt.update(value.into());
                    } else {
                        self.model.#field_name = Some(mongod::Historic::new(value.into()));
                    }

                    self.changeset.entry("$set".to_string()).or_insert(mongod::mongodb::bson::doc! {}.into()).as_document_mut().unwrap().insert(
                        #field_name_str.to_string(),
                        mongod::mongodb::bson::to_bson(&self.model.#field_name).unwrap(),
                    );

                    self
                }
            };
        }

        let documentation = format!("Set the value of `{field_name}`. If `Some(_)` it will be updated or removed if it is `None`");

        return quote! {
            #[doc = #documentation]
            pub fn #field_name(mut self, value: Option<#inner_field_type>) -> Self {
                let is_some = value.is_some();

                self.model.#field_name = value.into();

                if is_some {
                    self.changeset.entry("$set".to_string()).or_insert(mongod::mongodb::bson::doc! {}.into()).as_document_mut().unwrap().insert(
                        #field_name_str.to_string(),
                        mongod::mongodb::bson::to_bson(&self.model.#field_name).unwrap(),
                    );
                } else {
                    self.changeset.entry("$unset".to_string()).or_insert(mongod::mongodb::bson::doc! {}.into()).as_document_mut().unwrap().insert(
                        #field_name_str.to_string(),
                        mongod::mongodb::bson::to_bson("").unwrap(),
                    );
                }

                self
            }
        };
    }

    let documentation = format!("Set the value of `{field_name}`");
    // Code for T
    quote! {
        #[doc = #documentation]
        pub fn #field_name<T>(mut self, value: T)-> Self where T: Into<#field_type> + serde::Serialize  {
            self.model.#field_name = value.into();

            self.changeset.entry("$set".to_string()).or_insert(mongod::mongodb::bson::doc! {}.into()).as_document_mut().unwrap().insert(
                #field_name_str.to_string(),
                mongod::mongodb::bson::to_bson(&self.model.#field_name).unwrap(),
            );

            self
        }
    }
}
