extern crate proc_macro;
use case::CaseExt;

use proc_macro::TokenStream;
use quote::quote;
use syn::{parse_macro_input, Data, DeriveInput, Fields};

mod types;
use types::*;
mod partial_fields;
use partial_fields::partial_code_field;
mod update_fields;
use update_fields::update_code_field;
mod change_fields;
use change_fields::builder_change_fields;

/// #[derive(Model)]
#[proc_macro_derive(Model)]
pub fn model_derive(input: TokenStream) -> TokenStream {
    let input = parse_macro_input!(input as DeriveInput);

    let name = input.ident;
    let name_str = name.to_string().to_snake();
    let partial_name = syn::Ident::new(&format!("Partial{}", name), name.span());
    let changebuilder_name = syn::Ident::new(&format!("Change{}", name), name.span());

    let update_doc = "Commit the builder changes to DB";

    // Generate code for each field
    let field_code = if let Data::Struct(data_struct) = input.data {
        match data_struct.fields {
            Fields::Named(fields_named) => {
                // Update code
                let field_process_code: Vec<_> =
                    fields_named.named.iter().map(update_code_field).collect();

                // Partial struct fields
                let partial_struct: Vec<_> =
                    fields_named.named.iter().map(partial_code_field).collect();

                // Builder functions
                let builder_change_fields: Vec<_> = fields_named
                    .named
                    .iter()
                    .map(builder_change_fields)
                    .collect();

                quote! {
                    impl mongod::model::Model for #name {
                        type Partial = #partial_name;
                        type ChangeBuilder = #changebuilder_name;

                        fn change_builder(self) -> Self::ChangeBuilder {
                            #changebuilder_name::new(self)
                        }

                        async fn update_values(
                            &mut self,
                            obj: &serde_json::Map<String, serde_json::Value>,
                            update: &mut mongod::mongodb::bson::Document,
                        ) {
                            #( #field_process_code )*
                        }
                    }

                    #[derive(Debug, Clone, serde::Serialize, serde::Deserialize)]
                    pub struct #partial_name {
                        #( #partial_struct )*
                    }

                    #[derive(Debug)]
                    pub struct #changebuilder_name {
                        model: #name,
                        changeset: mongod::mongodb::bson::Document
                    }

                    impl #changebuilder_name {
                    pub fn new(model: #name) -> Self {
                        Self {
                            model,
                            changeset: mongod::mongodb::bson::doc! {}
                        }
                    }

                    #[doc = #update_doc]
                    pub async fn update(self) -> Result<#name, mongod::model::UpdateError> {
                            let db = mongod::get_mongo!();
                            let collection = mongod::col!(db, <#name as mongod::Referencable>::collection_name());

                            if let Err(msg) = mongod::Validate::validate(&self.model).await {
                                    return Err(mongod::model::UpdateError::Validation(msg));
                            }

                            let changeset = self.changeset;

                            collection
                                    .update_one(
                                        mongod::id_of!(mongod::Referencable::id(&self.model)),
                                        changeset,
                                        None,
                                    )
                                    .await
                                    .map_err(mongod::model::UpdateError::Database)?;

                                Ok(self.model)
                    }

                    #( #builder_change_fields )*

                    }

                    impl mongod::model::reference::Referencable for #partial_name {
                        fn collection_name() -> &'static str {
                            #name_str
                        }

                        fn id(&self) -> &str {
                            &self._id
                        }
                    }
                }
            }
            _ => unimplemented!(),
        }
    } else {
        unimplemented!()
    };

    TokenStream::from(field_code)
}

/// #[derive(Referencable)]
#[proc_macro_derive(Referencable)]
pub fn referencable_derive(input: TokenStream) -> TokenStream {
    let input = parse_macro_input!(input as DeriveInput);

    let name = input.ident;
    let name_str = name.to_string().to_snake();

    let code = quote! {
        impl mongod::model::reference::Referencable for #name {
            fn collection_name() -> &'static str {
                #name_str
            }

            fn id(&self) -> &str {
                &self._id
            }
        }
    };

    TokenStream::from(code)
}
